/* morse.c */


#define TEST

volatile unsigned int timecnt1=0;
volatile unsigned char n;

#define DEFAULT_T_DIT 	150

#define T_DIT  eeprom_read_byte(&ee_data.t_dit_ms)
#define T_DAH			4*T_DIT
#define T_INTER_CHAR	6*T_DIT
#define T_SPACE 		9*T_DIT

#define MORSE_TEXT_LEN	16

#include <ctype.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include "uart.h"


struct s_eeprom_data{
	unsigned char t_dit_ms;
	char	 text1[MORSE_TEXT_LEN];
}ee_data EEMEM;


struct s_morsecode{
 char c;
 unsigned char pattern;
}morsecode[] PROGMEM =
{	{'e', 0b10}, {'n', 0b110}, {'r', 0b1010}, {'s', 0b1000}, {'t', 0b11}, {'l',0b10100},
	{'1', 0b101111},
	{'2', 0b100111},
	{'3', 0b100011},
	{'4', 0b100001},
	{'5', 0b100000},
	{'6', 0b110000},
	{'7', 0b111000},
	{'8', 0b111100},
	{'9', 0b111110},
	{'0', 0b111111},
	{'a', 0b101}, {'b', 0b11000}, {'c', 0b11010}, {'d',0b1100}, {'f', 0b10010},
	{'g', 0b1110}, {'h', 0b10000}, {'i', 0b100}, {'j', 0b10111}, {'k', 0b1101},
	{'m', 0b111}, {'o', 0b1111}, {'p', 0b10110}, {'q', 0b11101}, {'u', 0b1001},
	{'v', 0b10001}, {'w', 0b1011}, {'x', 0b11001}, {'y',0b11011}, {'z', 0b11100},
	{'/', 0b110010}, {'+', 0b101010}, {'=',0b110001 }, {'?',0b1001100 }
};

#define NUM_MORSE_PATTERNS sizeof(morsecode)/sizeof(struct s_morsecode)


enum {OFF, ON};

#define MORSE_PORT PORTD
#define MORSE_DDR  DDRD
#define MORSE_BIT 6

#define PWM_PORT PORTD
#define PWM_DDR  DDRD
#define PWM_BIT 7

#define morse_bit(x) {if (x==ON) MORSE_PORT |= (1<<MORSE_BIT); else MORSE_PORT &= ~(1<<MORSE_BIT);}

void morse_init( unsigned char t_dit, char *p );
unsigned char get_morsecode (char ch);
unsigned char  morse( char ch );



ISR(TIMER0_OVF_vect) {

	// if output active, toggle every 500ms
	if (MORSE_PORT & (1<<MORSE_BIT)) // if output active
		PWM_PORT ^= (1<<PWM_BIT);


	// decrement millisecond timers
	if (++n & 1){	// 
		if (timecnt1 != 0){
			timecnt1--;
		}

	}

	return;
}


void set_timecnt(unsigned int t){

	while(timecnt1);
	cli();
	timecnt1 = t;
	sei();
	return;
}

void morse_init( unsigned char dittime, char *p){

	unsigned char i;

	if (T_DIT != dittime){
		if ((dittime > 20) && (dittime < 250)){
			eeprom_write_byte(&ee_data.t_dit_ms, dittime);
		}
		else{
			eeprom_write_byte(&ee_data.t_dit_ms, DEFAULT_T_DIT);
		}
	}

	for (i=0; i<MORSE_TEXT_LEN; i++){
		if (eeprom_read_byte(&ee_data.text1[i]) != *p){
			eeprom_write_byte(&ee_data.text1[i], *p);
		}
		if (*p==0){
			 i=MORSE_TEXT_LEN;
		}
		else{
			p++;
		}
	}

	morse_bit(OFF);
	MORSE_DDR |= (1<<MORSE_BIT);
	PWM_DDR   |= (1<<PWM_BIT);

	// init timer0
	TCCR0 = (1<<CS01); // divider 8
	TIMSK |= (1<<TOIE0);

	return;
}

unsigned char get_morsecode (char ch){
unsigned char i, x;

	x = i = 0;
	ch = tolower(ch);

	do{	if (ch == pgm_read_byte(&morsecode[i].c)){
			x = pgm_read_byte(&morsecode[i].pattern);
			i=NUM_MORSE_PATTERNS;
		}
	} while (++i < NUM_MORSE_PATTERNS);

	return(x);
}


unsigned char morse( char ch ){
	unsigned char mcode, mask;

	if (isspace(ch)){
		timecnt1 = T_SPACE;
		while(timecnt1)
			;	// decremented by interrupt
		return(0);
	}

	mcode = get_morsecode(ch);

	// send out morse pattern
	mask = 0x80;
	// find start bit
	while ((mcode & mask) == 0)
	  mask >>= 1;	
	// skip start bit
	mask >>=1;

// short delay for testing purposes
	set_timecnt(20);
	while(timecnt1)
		;	// decremented by interrupt
	
	while (mask){
		if ((mcode & mask) == 0)
			set_timecnt(T_DIT);
		else
			set_timecnt(T_DAH);

		morse_bit(ON);
		while(timecnt1)
			;	// decremented by interrupt


		// inter symbol space
		morse_bit(OFF);
		set_timecnt(3*T_DIT/2);
		while(timecnt1)
			;	// decremented by interrupt

		// shift mask to next bit
		mask >>=1;	
	}


	// inter character space
	set_timecnt(T_INTER_CHAR-T_DIT);
	while(timecnt1)
			;	// decremented by interrupt


	return(mcode);
}


#ifdef TEST
int main(void){

	char txt[] = "73 de df2ds";
	char i, cc;
	unsigned char code;
	unsigned int res;


	uart_init(UART_BAUD_SELECT(9600, 4000000L));


	morse_init(90, &txt[0]);
	sei();

	while(1){

// for testing
		do{
			res = uart_getc();
		}while (res == UART_NO_DATA);
		cc = res & 0xFF;
		uart_putc(cc); uart_putc(' ');
		code = get_morsecode(cc);

		for (i=0x80; i!=0; i>>=1){
			if (i & code) uart_putc('1');
			else uart_putc('0');
		}

		uart_puts("\n\r");
		morse(cc);
	}

	return(0);
}
#endif
